/*
*****************************************************************************
*
*                          Copyright (c) 2000 2005
*
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice, 
*     this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright notice, 
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*  - The names of its contributors may not be used to endorse or promote 
*     products derived from this software without specific prior written 
*     permission.
*   
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*  POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************
* File          : loop.c
* Author        : Zadig
* Creation date : 27/01/2005
*.............................................................................
* Description :
******************************************************************************
*/

/* ########################################################################## */
/* ### #include                                                               */
#include <drivers/Drivers.h>

#include "loop.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>

/* ########################################################################## */
/* ### Functions                                                              */
status_t open_hook(const char *name, uint32 flags, void **cookie);
status_t close_hook(void *cookie);
status_t free_hook(void *cookie);
status_t read_hook(void *cookie, off_t position, void *data, size_t *len);
status_t write_hook(void *cookie, off_t position, const void *data, size_t *len);
status_t control_hook(void *cookie, uint32 op, void *data, size_t len);

/* ########################################################################## */
/* ### Variables                                                              */
int32 api_version = B_CUR_DRIVER_API_VERSION;

TS_loop_device gas_loop_Devices[LOOP_DEVICE_MAX];

struct cryptlib_module_info  *gps_CryptMod = NULL;

device_hooks dh = 
{
   open_hook, 
   close_hook, 
   free_hook,
   control_hook, 
   read_hook, 
   write_hook,
   NULL, NULL, NULL, NULL
};


/***************
 *    Log      *
 ***************/
#ifdef FLG_DBG
char     gac_loop_LogArray[LOG_SIZE];
uint32   gi_loop_LogSize=0;

status_t LOG_Add(char *psz_String)
{
   if( (gi_loop_LogSize + strlen(psz_String) + 1) < LOG_SIZE)
   {
      sprintf(&(gac_loop_LogArray[gi_loop_LogSize]), "%s", psz_String);
      gi_loop_LogSize += strlen(psz_String); /* do not include '\0' */
   }
   return(B_OK);
}
#endif


/************************************************************
 *  Function :
 *  Date     : 18/01/2005
 ************************************************************
 * Usage:
 * Called when the system is booted, to let the driver
 * detect and reset the hardware.
 ************************************************************/
status_t init_hardware(void)
{
   return(B_OK);
}

/************************************************************
 *  Function :
 *  Date     : 18/01/2005
 ************************************************************
 * Usage:
 * Called when the driver is loaded, so it can allocate
 * needed system resources.
 ************************************************************/
status_t init_driver(void)
{
   /* load crypt module */
   get_module(CRYPTLIB_MODULE_NAME, (module_info **) &gps_CryptMod);
   if(gps_CryptMod == NULL)
   {
      DBG_PRINT("unable to load cryptmod!");
      return(B_ERROR);
   }
 
   /* init device infos */
   if(LOOP_InitDevice(gas_loop_Devices, LOOP_DEVICE_MAX) != B_OK)
      return(B_ERROR);
   
   return(B_OK);
}

/************************************************************
 *  Function :
 *  Date     : 18/05/2005
 ************************************************************
 * Usage:
 * Called just before the driver is unloaded, so it can
 * free allocated resources.
 ************************************************************/
void uninit_driver(void)
{
   LOOP_FiniDevice(gas_loop_Devices, LOOP_DEVICE_MAX);
   put_module(CRYPTLIB_MODULE_NAME);
}

/************************************************************
 *  Function :
 *  Date     : 18/05/2005
 ************************************************************
 * Usage:
 * Called to obtain a list of device names supported by the
 * driver.
 ************************************************************/
const char **publish_devices(void)
{
   const static char *gas_DeviceList[LOOP_DEVICE_MAX + 1];
   uint16 i;
   
   for (i=0; i<LOOP_DEVICE_MAX; i++)
      gas_DeviceList[i] = gas_loop_Devices[i].name;
   
   gas_DeviceList[i] = 0;
   return (gas_DeviceList);
}

/************************************************************
 *  Function :
 *  Date     : 18/01/2005
 ************************************************************
 * Usage:
 * Called to obtain a list of pointers to the hook
 * functions for a specified device.
 ************************************************************/
device_hooks *find_device(const char *name)
{
   return (&dh);
}

/************************************************************
 *  Function :
 *  Date     : 18/01/2005
 ************************************************************
 * Usage:
 * Open function
 ************************************************************/
status_t open_hook(const char *name, uint32 flags, void **cookie)
{
   uint8 i;
   TS_loop_device *ps_Device;
   TS_loop_filedesc  *ps_Desc;
#ifdef FLG_DBG
   char  sz_LogString[LOOP_STRING_SIZE];
   
   sprintf(sz_LogString, 
            "open_hook flags=0x%X name='%s'\n", (int)flags, name);
   sz_LogString[LOOP_STRING_SIZE-1] = '\0';
   LOG_Add(sz_LogString);
#endif
  
   *cookie = NULL;
   /* find device and check that it is not already open */
   ps_Device = NULL;
   for(i=0; i<LOOP_DEVICE_MAX; i++)
   {
      if(strcmp(gas_loop_Devices[i].name, name) == 0)
      {
         ps_Device = &(gas_loop_Devices[i]);
         if(ps_Device->type == LOOP_DEVICE)
         {
            if(ps_Device->link[0] == '\0')  /* no link*/
               return(B_ERROR);
            else
               break;
         }
         else /* loop manager */
         {
            break;
         }
      }
   }
   if(i >= LOOP_DEVICE_MAX)
   {
      dprintf("Loop Device : device %s not found!\n", name);
      return(B_NAME_NOT_FOUND);
   }     
   
   ps_Desc = LOOP_NewDesc();
   if(ps_Desc == NULL)
   {
      *cookie = NULL;
      DBG_PRINT("no free descriptor to open file\n");
      return(B_NO_MEMORY);
   }
   ps_Desc->device = ps_Device;
   
   acquire_sem(ps_Device->sem_protect);
   if(ps_Device->type == LOOP_DEVICE)
   {
      ps_Desc->file = open(ps_Device->link, flags);
      if(ps_Desc->file >= 0)
      {
         /* get devices specs at first open */
         if(ps_Device->nb_open == 0)
            LOOP_SetDeviceGeometry(ps_Device, ps_Desc->file);
            
         if(ps_Device->dev_type == LOOP_DEVICE_FILE)
            ioctl(ps_Desc->file, IOCTL_FILE_UNCACHED_IO, NULL);
         
         if(LOOP_CryptInit(ps_Desc) != B_OK)
         {
            release_sem(ps_Device->sem_protect);
            LOOP_FreeDesc(ps_Desc);
            return(B_ERROR);
         }

         *cookie = ps_Desc;
         ps_Device->nb_open++;
         release_sem(ps_Device->sem_protect);
         return(B_OK);
      }
   }
   else if(ps_Device->type == LOOP_MANAGER)
   {
      *cookie = ps_Desc;
      ps_Device->nb_open++;;
      release_sem(ps_Device->sem_protect);
      return(B_OK);
   }
   release_sem(ps_Device->sem_protect);
   LOOP_FreeDesc(ps_Desc);
   return(B_ERROR);
}

/************************************************************
 *  Function :
 *  Date     : 18/01/2005
 ************************************************************
 * Usage:
 * Close function
 ************************************************************/
status_t close_hook(void *cookie)
{
   TS_loop_device *ps_Device;
   TS_loop_filedesc  *ps_Desc;
#ifdef FLG_DBG
   char  sz_LogString[LOOP_STRING_SIZE];

   sprintf(sz_LogString, "close_hook\n");
   sz_LogString[LOOP_STRING_SIZE-1] = '\0';
   LOG_Add(sz_LogString);
#endif

   ps_Desc = cookie;   
   ps_Device = ps_Desc->device;

   if(ps_Device->type == LOOP_DEVICE)
      close(ps_Desc->file);
   
   return(B_OK);
}

/************************************************************
 *  Function :
 *  Date     : 18/01/2005
 ************************************************************
 * Usage:
 * free  function
 ************************************************************/
status_t free_hook(void *cookie)
{
   TS_loop_filedesc  *ps_Desc;
   TS_loop_device *ps_Device;
   
   ps_Desc = cookie;   
   ps_Device = ps_Desc->device;
   
   acquire_sem(ps_Device->sem_protect);
   ps_Device->nb_open--;
   if(ps_Device->type == LOOP_DEVICE)
      LOOP_CryptFini(ps_Desc);
   
   LOOP_FreeDesc(ps_Desc);
   release_sem(ps_Device->sem_protect);
   return(B_OK);
}

/************************************************************
 *  Function :
 *  Date     : 18/01/2005
 ************************************************************
 * Usage:
 * Read  function
 ************************************************************/
status_t read_hook(void *cookie, off_t position, void *data, size_t *len)
{
#ifdef FLG_DBG
   char  sz_LogString[LOOP_STRING_SIZE];
#endif
   int i_ret;
   TS_loop_device *ps_Device;
   TS_loop_filedesc  *ps_Desc;

   ps_Desc = cookie;
   ps_Device = ps_Desc->device;
   if(ps_Device->type == LOOP_DEVICE)
   {
      uint32 i_Size, i_RetLen;

      /* we handle only request multiple of block_size */
      if( ((*len) % ps_Device->block_size != 0) ||
          (position % ps_Device->block_size != 0) )
      {
#ifdef FLG_DBG
      sprintf(sz_LogString, "read_hook pos=0x%X len=0x%X\n", 
               (int)position, (int)*len);
      sz_LogString[LOOP_STRING_SIZE-1] = '\0';
      LOG_Add(sz_LogString);
#endif
         *len = 0;
         return(B_ERROR);
      }
      
      /* read the datas */
      i_RetLen = 0;

      i_ret = read_pos(ps_Desc->file, position, data, *len);
      if(i_ret != *len)
      {
#ifdef FLG_DBG
         sprintf(sz_LogString, "read_pos error 0x%X\n", i_ret);
         sz_LogString[LOOP_STRING_SIZE-1] = '\0';
         LOG_Add(sz_LogString);
#endif
         *len = 0;
         return(B_ERROR);
      }
      i_Size = i_ret;
      *len = i_ret;
      
      acquire_sem(ps_Desc->sem_protect);
      while(i_Size > 0)
      {
         /* decrypt data */
         if(LOOP_Decrypt(ps_Desc, position, data) != B_OK)
         {
#ifdef FLG_DBG
            sprintf(sz_LogString, "decrypt error\n");
            sz_LogString[LOOP_STRING_SIZE-1] = '\0';
            LOG_Add(sz_LogString);
#endif
            *len = i_RetLen;
            release_sem(ps_Desc->sem_protect);
            return(B_OK);
         }

         /* goto next block */
         i_RetLen += ps_Device->block_size;
         position += ps_Device->block_size;
         data += ps_Device->block_size;
         i_Size -= ps_Device->block_size;
      }
      release_sem(ps_Desc->sem_protect);
   }
   else  /* manager */
   {
#ifdef FLG_DBG
      if (position > gi_loop_LogSize)
      {
         *len = 0;
      }
      else
      {
         uint32 i_NumRead=0;
         
         while( (position < gi_loop_LogSize) && (i_NumRead < *len))
         {
            *(char *)data++ = gac_loop_LogArray[position++];
            i_NumRead++;
         }
         *len = i_NumRead;
      }
#else
      return(B_PERMISSION_DENIED);
#endif
   }
   return(B_OK);
}

/************************************************************
 *  Function :
 *  Date     : 18/01/2005
 ************************************************************
 * Usage:
 * Write function.
 * The write buffer can not be used to crypt the datas as it is done in the read
 * hook because it can be used by other threads when we write it. Thus a
 * temporary buffer is used.
 * Malloc is not a good idea in a driver but this allow to do only one device
 * access per write call (since physical accesses are slow this is very
 * important). This can not be ensured this static buffers.
 ************************************************************/
status_t write_hook(void *cookie, off_t position, const void *data, size_t *len)
{
   int32 i_ret;
   TS_loop_device *ps_Device;
   TS_loop_filedesc  *ps_Desc;
#ifdef FLG_DBG
   char  sz_LogString[LOOP_STRING_SIZE];
#endif

   ps_Desc = cookie;
   ps_Device = ps_Desc->device;
   
   if(ps_Device->type == LOOP_DEVICE)
   {
      uint8 *p_TempBuffer, *p_CurBuffer;
      uint32 i_Size;
      uint32 i_CurPosition;
      
      /* we handle only request multiple of block_size */
      if( ((*len) % ps_Device->block_size != 0) ||
          (position % ps_Device->block_size != 0) )
      {
#ifdef FLG_DBG
         sprintf(sz_LogString, "write_hook pos=0x%X len=0x%X\n", 
                  (int)position, (int)*len);
         sz_LogString[LOOP_STRING_SIZE-1] = '\0';
         LOG_Add(sz_LogString);
#endif
         *len = 0;
         return(B_ERROR);
      }
      
      p_TempBuffer = (uint8*)malloc(*len);
      if(p_TempBuffer == NULL)
      {
         *len = 0;
         DBG_PRINT("write malloc error");
         return(B_ERROR);
      }

      /* copy the datas to the temp buffer */
      i_Size = *len;
      memcpy(p_TempBuffer, data, *len);
      p_CurBuffer = p_TempBuffer;
      i_CurPosition = position;
      
      /* crypt data */
      acquire_sem(ps_Desc->sem_protect);
      while(i_Size > 0)
      {
         if(LOOP_Crypt(ps_Desc, i_CurPosition, p_CurBuffer) != B_OK)
         {
            *len = 0;
            release_sem(ps_Desc->sem_protect);
            free(p_TempBuffer);
            return(B_ERROR);
         }
         
         /* goto next block */
         i_CurPosition += ps_Device->block_size;
         p_CurBuffer += ps_Device->block_size;
         i_Size -= ps_Device->block_size;
      }
      release_sem(ps_Desc->sem_protect);

      i_ret = write_pos(ps_Desc->file, position, p_TempBuffer, *len);
      if(i_ret < 0)
      {
         *len = 0;
         free(p_TempBuffer);
         return(B_ERROR);            
      }

      *len = i_ret;
      free(p_TempBuffer);
      return(B_OK);
   }
   else
      return(B_PERMISSION_DENIED);
}

/************************************************************
 *  Function :
 *  Date     : 18/01/2005
 ************************************************************
 * Usage:
 * ioctl function
 ************************************************************/
status_t control_hook(void *cookie, uint32 op, void *data, size_t len)
{
#ifdef FLG_DBG
   char  sz_LogString[LOOP_STRING_SIZE];
#endif
   status_t i_ret = B_OK;
   TS_loop_device *ps_Device;
   TS_loop_filedesc  *ps_Desc;


   ps_Desc = cookie;
   ps_Device = ps_Desc->device;
   
   if(ps_Device->type == LOOP_DEVICE)
   {
      switch(op)
      {
         case B_GET_DEVICE_SIZE:
            /* truncated... */
            if(ps_Device->dev_type == LOOP_DEVICE_DEV)
            {
               i_ret = ioctl(ps_Desc->file, op, data);
            }
            else
            {
               *(size_t *)data = (size_t)((off_t)ps_Device->geometry.bytes_per_sector * 
                     ps_Device->geometry.sectors_per_track * 
                     ps_Device->geometry.cylinder_count *
                     ps_Device->geometry.head_count);
            }
            break;
            
         case B_GET_MEDIA_STATUS:
            if(ps_Device->dev_type == LOOP_DEVICE_DEV)
               i_ret = ioctl(ps_Desc->file, op, data);
            else
               *(status_t *)data = B_OK;
            break;
           
         case B_GET_BIOS_GEOMETRY:
         case B_GET_GEOMETRY:
            if(ps_Device->dev_type == LOOP_DEVICE_DEV)
            {
               i_ret = ioctl(ps_Desc->file, op, data);
            }
            else
            {
               len = sizeof(device_geometry);
               memcpy(data, &ps_Device->geometry, len);
            }
            break;            
         
         case B_GET_ICON:
            i_ret = LOOP_GetIcon((device_icon *)data);
            break;
      
            /* device only ioctls */
         case B_FLUSH_DRIVE_CACHE:
            if(ps_Device->dev_type == LOOP_DEVICE_DEV)
            {
               i_ret = ioctl(ps_Desc->file, op, data);
            }
            else
            {
               fsync(ps_Desc->file); /* this is asynchronous: pb ? */
               *(status_t*)data = B_OK;
            }
            break;
         
         case B_FORMAT_DEVICE: /* low level format */
            if(ps_Device->dev_type == LOOP_DEVICE_DEV)
               i_ret = ioctl(ps_Desc->file, op, data);
            break;
            
         case B_EJECT_DEVICE:
            if(ps_Device->dev_type == LOOP_DEVICE_DEV)
               i_ret = ioctl(ps_Desc->file, op, data);
            else
               i_ret = B_ERROR;
            break;
            
            /* unsupported ioctls */
         case B_GET_NEXT_OPEN_DEVICE: 
         case B_GET_DRIVER_FOR_DEVICE:
         default:
            i_ret = B_ERROR;
            break;
      }

#ifdef FLG_DBG
      sprintf(sz_LogString, "control_hook op=0x%X len=0x%X ret= 0x%X\n", 
               (int)op, (int)len, (int)i_ret);
      sz_LogString[LOOP_STRING_SIZE-1] = '\0';
      DBG_PRINT(sz_LogString);
      LOG_Add(sz_LogString);
#endif
      return(i_ret);
   }
   else if(ps_Device->type == LOOP_MANAGER)
   {
      TS_LOOP_cmd *ps_Cmd = data;
      TS_loop_device *ps_LoopDevice;
    
      if(ps_Cmd == NULL)
         return(B_BAD_VALUE);
      
      /* find target loop and check ioctl */
      switch(op)
      {
         case LOOP_SET_LOOP:
         case LOOP_GET_LOOP:
         case LOOP_SET_ALGO:
         case LOOP_GET_ALGO:
         case LOOP_SET_KEY:
         case LOOP_GET_KEY:
            ps_LoopDevice = LOOP_FindDevice(gas_loop_Devices, 
                                             LOOP_DEVICE_MAX, ps_Cmd->loop);
            if(ps_LoopDevice == NULL)
            {
               dprintf("Loop Device : ioctl/ unable to find device %s!\n",
                     ps_Cmd->loop);
               return(B_ERROR);
            }
            break;
         default:
            return(B_ERROR);
      }

      switch(op)
      {
         case LOOP_SET_LOOP:
            return(LOOP_ManSetLoop(ps_LoopDevice, &(ps_Cmd->data.loop)));
            break;
         case LOOP_GET_LOOP:
            return(LOOP_ManGetLoop(ps_LoopDevice, &(ps_Cmd->data.loop)));
            break;

         case LOOP_SET_ALGO:
            return(LOOP_ManSetAlgo(ps_LoopDevice, &(ps_Cmd->data.algo)));
            break;
         
         case LOOP_GET_ALGO:
            return(LOOP_ManGetAlgo(ps_LoopDevice, &(ps_Cmd->data.algo)));
            break;

         case LOOP_SET_KEY:
            return(LOOP_ManSetKey(ps_LoopDevice, &(ps_Cmd->data.key)));
            break;

         case LOOP_GET_KEY:
            return(LOOP_ManGetKey(ps_LoopDevice, &(ps_Cmd->data.key)));
            break;

         default:
            return(B_ERROR);
            break;
      }
   }
   return(B_OK);
}

